/* 

dvbtextsubs - a teletext subtitles decoder for DVB cards. 
(C) Dave Chapman <dave@dchapman.com> 2003-2004.

The latest version can be found at http://www.linuxstb.org/

Thanks to:  

Ralph Metzler (re: dvbtext - dvbtextsubs is based heavily on dvbtext)
for his work on both the DVB driver and his old vbidecode package
(some code and ideas in dvbtext are borrowed from vbidecode).

Jan Pantelje for his advice and his work on submux-dvd

Ragnar Sundblad (the author of the VDR teletext subtitles plugin) for
his help in adding VDR support to dvbtextsubs and for suggesting
various improvements.

Scott T. Smith for creating "dvdauthor".

Copyright notice:

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

// Linux includes:
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>

// DVB includes:
#ifndef WITHOUT_DVB
  #include <linux/dvb/dmx.h>
  #include <linux/dvb/frontend.h>
#endif

// dvbtext includes:
#include "tables.h"
#ifndef WITHOUT_DVB
  #include "tune.h"
#endif
#include "vtxdecode.h"

#define VERSION "0.2-srt"
#define USAGE "\nUSAGE: dvbtextsubs [-r] [-o offset] [-f FREQ -p POL -s SR] pid pageno\nor     dvbtextsubs -vdr pageno\n\nUse -f, -p -s to specify a satellite transponder\nor -vdr to read a .vdr file from stdin\notherwise, the input is a transport stream from stdin.\n\n"

int debug=0;
int no_pts_warning=0;

/* Structure of Transport Stream (from ISO/IEC 13818-1):

transport_packet() {
  sync_byte                                   8    bslbf       0
  transport_error_indicator                   1    bslbf       1
  payload_unit_start_indicator                1    bslbf       1
  transport_priority                          1    bslbf       1
  PID                                         13   uimsbf      1,2
  transport_scrambling_control                2    bslbf       3
  adaption_field_control                      2    bslbf       3
  continuity_counter                          4    uimsbf      3
  if (adaption_field_control=='10' || adaption_field_control=='11'){
      adaption_field()
  }
  if (adaption_field_control=='01' || adaption_field_control=='11'){
      for (i=0;i<N;i++){
        data_byte                             8    bslbf
      }
  }
}

*/

int verbose=0;
int displayed_header=0;
uint64_t PTS=0;
uint64_t audio_pts=0;
uint64_t first_audio_pts=0;
uint64_t video_pts=0;
uint64_t first_video_pts=0;
uint64_t FIRST_PTS=0;

typedef struct {
  unsigned char lang;
  uint64_t start_PTS;
  uint64_t end_PTS;
  int has_end_pts;
  int num_lines;
  int row_no[25];
  char lines[25][40];
  int first_char[25];
  int last_char[25];
} subtitle_t;

typedef enum {
  TS_WAITING,
  TS_ERROR,
  TS_IN_PAYLOAD
} ts_status_t;

int srt_flag;
double offset;

int check_pes(unsigned char* buf,int n) {
  int i=0;
  int stream_id;
  int PES_packet_length;

  if ((buf[0]!=0) || (buf[1]!=0) || (buf[2]!=1)) {
    fprintf(stderr,"PES ERROR: does not start with 0x000001\n");
    return(0);
//  } else {
//    fprintf(stderr,"PES starts with 0x000001\n");
  }
  i=3;
  // stream_id: e0=video, c0=audio, bd=DVB subtitles, AC3 etc
  stream_id=buf[i++];

  PES_packet_length=(buf[i]<<8)|buf[i+1]; i+=2;
//fprintf(stderr,"PES_packet_length=%d\n",PES_packet_length);

  if ((stream_id&0xe0)==0xe0) {
    // Video stream - PES_packet_length must be equal to zero
    if (PES_packet_length!=0) {
      fprintf(stderr,"ERROR IN VIDEO PES STREAM: PES_packet_length=%d (must be zero)\n",PES_packet_length);
      return(0);
    }
  } else {
    if (PES_packet_length+6!=n) {
      fprintf(stderr,"ERROR IN NON-VIDEO PES: n=%d,PES_packet_length=%d\n",n,PES_packet_length);
      return(0);
    }
  }
  return(1);
}

static const char* colours[]= {
  "black","red","lime","yellow","blue","magenta","cyan","white"
};

// There seems to be a limit of 8 teletext streams - OK for most (but
// not all) transponders.
#define MAX_CHANNELS 8

typedef struct mag_struct_ {
   int valid;
   int mag;
   unsigned char flags;
   unsigned char lang;
   uint64_t PTS;
   int line_transmitted[25];
   int num_valid_lines;
   int pnum,sub;
   unsigned char pagebuf[25*40];
} mag_struct;

  mag_struct prevpage;
  mag_struct thepage;

// FROM vbidecode
// unham 2 bytes into 1, report 2 bit errors but ignore them
unsigned char unham(unsigned char a,unsigned char b)
{
  unsigned char c1,c2;
  
  c1=unhamtab[a];
  c2=unhamtab[b];
//  if ((c1|c2)&0x40) 
//      fprintf(stderr,"bad ham!");
  return (c2<<4)|(0x0f&c1);
}

void xml_output_char(FILE* fp, unsigned char ch) {

  switch(ch) {
    case '<': fprintf(fp,"&lt;"); break;
    case '>': fprintf(fp,"&gt;"); break;
    case '&': fprintf(fp,"&amp;"); break;
//  case '"': fprintf(fp,"&quote;"); break;
    default:  fprintf(fp,"%c",ch); break;
  }
}

#ifndef WITHOUT_DVB
void set_tt_filt(int fd,uint16_t tt_pid)
{
	struct dmx_pes_filter_params pesFilterParams;

	pesFilterParams.pid     = tt_pid;
	pesFilterParams.input   = DMX_IN_FRONTEND;
	pesFilterParams.output  = DMX_OUT_TS_TAP;
        pesFilterParams.pes_type = DMX_PES_TELETEXT;
	pesFilterParams.flags   = DMX_IMMEDIATE_START;

	if (ioctl(fd, DMX_SET_PES_FILTER, &pesFilterParams) < 0)  {
                fprintf(stderr,"FILTER %i: ",tt_pid);
		perror("DMX SET PES FILTER");
        }
}
#endif

char pts_text[30];
char* pts2hmsu(uint64_t pts) {
  int h,m,s,u;

  h=(pts/(1000*60*60));
  m=(pts/(1000*60))-(h*60);
  s=(pts/1000)-(h*3600)-(m*60);
  u=pts-(h*1000*60*60)-(m*1000*60)-(s*1000);

  sprintf(pts_text,"%d:%02d:%02d.%03d",h,m,s,u);
  return(pts_text);
}

char* pts2srt(uint64_t pts)
{
int h, m, s, u;

pts += offset * 1000.0;

h = (pts / (1000 * 60 * 60));
m = (pts / (1000 * 60) ) - (h * 60);
s = (pts / 1000) - (h * 3600) - (m * 60);
u = pts - (h * 1000 * 60 * 60) - (m * 1000 * 60) - (s * 1000);

sprintf(pts_text, "%02d:%02d:%02d,%03d", h, m, s, u);

return(pts_text);
} /* end function pts2srt */


int is_first=1;
subtitle_t subtitle;
subtitle_t prev_subtitle;

void print_xml(FILE* fd, subtitle_t* subtitle) {
  int i;
  char ch;
  int row;
  char tmp[41];
  int n;
  int colour;

  if (subtitle->num_lines==0) { return; }

//  printf("\"");
//  for (i=1;i<=40;i++) printf("%c",'0'+(i%10));
//  printf("\"\n");

  for (row=0;row<subtitle->num_lines;row++) {
    subtitle->first_char[row]=-1;
    i=0;
    n=0;
    for (i=1;i<=40;i++) {
      ch=subtitle->lines[row][i-1]&0x7f;
      if ((subtitle->first_char[row]<0) && (ch > 32)) subtitle->first_char[row]=i;
      if (ch < 32) ch=32;
      tmp[n++]=ch;
    }
    subtitle->last_char[row]=40;
    while ((subtitle->last_char[row] >0) && (tmp[subtitle->last_char[row]-1]==' ')) subtitle->last_char[row]--;
    tmp[n]=0;
//    fprintf(fd,"\"%s\"\n",tmp);
//    fprintf(fd,"First_char[%d]=%d, last_char[%d]=%d\n",row,subtitle->first_char[row],row,subtitle->last_char[row]);
  }

  fprintf(fd,"    <spu lang=\"%s\" start=\"%s\"",langs[subtitle->lang],pts2hmsu(subtitle->start_PTS));
  if (subtitle->has_end_pts) {
    fprintf(fd," end=\"%s\"",pts2hmsu(subtitle->end_PTS));
  }
  fprintf(fd,">\n");
  
  for (row=0;row<subtitle->num_lines;row++) {
    fprintf(fd,"      <line row=\"%d\" col=\"%d\">",subtitle->row_no[row],subtitle->first_char[row]);
    colour=-1;
    for (i=1;i<=40;i++) {
      ch=subtitle->lines[row][i-1]&0x7f;
      if (ch >= ' ') {
        if ((i >= subtitle->first_char[row]) && (i <= subtitle->last_char[row])) {
          if (colour==-1) {
            colour=7; fprintf(fd,"<%s>",colours[colour]);
          }
          xml_output_char(fd,vtx2iso8559_1_table[subtitle->lang][ch-32]);
        }
      } else {
        if ((i >= subtitle->first_char[row]) && (i <= subtitle->last_char[row])) fprintf(fd," ");
        if (ch < 8) {
          if (colour!=-1) {
            fprintf(fd,"</%s>",colours[colour]);
          }
          colour=ch;
          fprintf(fd,"<%s>",colours[colour]);
        } else if (ch == 29) {
          fprintf(fd,"<bg />");
        } else if ((ch!=0x0b) && (ch!=0x0a) &&   // Start/end box
                   (ch!=0x0d)) {                 // Double height
          fprintf(fd,"<%02x>",ch);          
        }
      }
    }
    if (colour!=-1) fprintf(fd,"</%s>",colours[colour]);
    fprintf(fd,"</line>\n");
  }
  fprintf(fd,"    </spu>\n");
}


void print_srt(FILE* fd, subtitle_t* subtitle)
{
int i;
char ch;
int row;
int n;
static char start_time[80];
static char end_time[80];
static char text[25][41];
static int have_sub_flag;
static int have_end_time_flag;
static int sub_number;
int sub_lines;
int leading_space_flag;

if(debug)
	{
	fprintf(stderr, "print_srt(fd=%p subtitle=%p\n", fd, subtitle);
	}

if(! fd) return;
if(! subtitle) return;

if(subtitle->num_lines == 0) return;

if(have_sub_flag)
	{
	if(! have_end_time_flag)
		{
		sprintf(end_time, pts2srt(subtitle->start_PTS) );
		}

	fprintf(fd, "%d\r\n%s --> %s\r\n", sub_number, start_time, end_time);

	for(i = 0; i < sub_lines; i++)
		{
		fprintf(fd, "%s\r\n", text[i]);
		}

	fprintf(fd, "\r\n");

	have_sub_flag = 0;
	}

sub_lines = 0;
for(row = 0; row < subtitle->num_lines; row++)
	{
	text[sub_lines][0] = 0;

    subtitle->first_char[row] = -1;
    i = 0;
    n = 0;

	/* for all characters in a row */
	leading_space_flag = 1;	
    for(i = 1; i <= 40; i++)
		{
        ch = subtitle->lines[row][i - 1] & 0x7f;
        if( (subtitle->first_char[row] < 0) && (ch > 32)) subtitle->first_char[row] = i;

		/* crtrl chars become spaces */
       	if(ch < 32) ch = 32;

		if(ch != 32) leading_space_flag = 0;

		if(! leading_space_flag)
			{
			/* accumulate line characters in tmp; */
    	   	text[sub_lines][n] = ch;

			n++;
			}
		}

	/* string termination */
	text[sub_lines][n] = 0;

	subtitle->last_char[row] = 40;

    while((subtitle->last_char[row] > 0) && (text[sub_lines][subtitle->last_char[row] - 1] == ' '))
		{
		subtitle->last_char[row]--;
		}

	sub_lines++;
	} /* end for all lines */

sprintf(start_time, "%s", pts2srt(subtitle->start_PTS) );

if(subtitle->has_end_pts)
	{
	sprintf(end_time, pts2srt(subtitle->end_PTS) );
	have_end_time_flag = 1;
	}
else have_end_time_flag = 0;

sub_number++;

have_sub_flag = 1;

}  /* end function print_srt */


void print_page(mag_struct *mag) {
  int line,i,j,flag;

  subtitle.num_lines=0;
  subtitle.start_PTS=mag->PTS-FIRST_PTS;
  subtitle.lang=mag->lang;

  if (mag->num_valid_lines>0) {
    for (i=1;i<=23;i++) {
      if (mag->line_transmitted[i]) {
        subtitle.row_no[subtitle.num_lines]=i;
        memcpy(subtitle.lines[subtitle.num_lines],&mag->pagebuf[40*i],40);
        subtitle.num_lines++;
      }
    }
  }
  if (!is_first) {
    if (subtitle.num_lines==0) {
      prev_subtitle.has_end_pts=1;
      prev_subtitle.end_PTS=subtitle.start_PTS;
    } else {
      prev_subtitle.has_end_pts=0;
    }
	if(! srt_flag)
		{
	    print_xml(stdout, &prev_subtitle);
		}
	else
		{
		print_srt(stdout, &prev_subtitle);
		}

    prev_subtitle=subtitle;
  }
  is_first=0;
}


void debug_print_page(mag_struct *mag) {
  int line,i,j,flag;
  unsigned char ch;
  int pageflags;
  pageflags=unham(mag->pagebuf[2],mag->pagebuf[3])&0x80;

  fprintf(stdout,"<page pageno=\"%03x\" flags=\"%02x\" charset=\"%s\" pts=\"%lld\" time=\"%.3f\">\n",mag->pnum,pageflags,langs[mag->lang],mag->PTS,(mag->PTS-FIRST_PTS)/1000.0);
  for (i=0;i<=23;i++) {
    if (mag->line_transmitted[i]) {
      flag=0;
      printf("%02d: ",i);
      for (j=0;j<39;j++) printf("%02x ",mag->pagebuf[40*i+j]);
      for (j=0;j<39;j++) {
        ch=(mag->pagebuf[40*i+j]&0x7f);
        if ((ch >=32) && (ch<=127)) { printf("%c",ch); } else { printf("."); }
      }
      printf("\n");
    }
  }

  fprintf(stdout,"</page>\n");  
}

void set_line(int line, unsigned char* data,int mag, int the_page) {
  unsigned char c;
  FILE * fd;
  char fname[80];
  unsigned char buf;
  int i;

  //  fprintf(stderr,"the_page=%04x\n",the_page);
  //  fprintf(stdout,"In set_line - mag=%d, line=%d!\n",mag,line);
  //  if (line==0) { fprintf(stderr,"mag=%d, page=%02x\n",mag,unham(data[0],data[1])); }
  if (mag!=((the_page&0x0f00)>>8)) {
     return;
  }
//    for (i=0;i<44;i++) {
//      if (((data[i]&0x7f) >= 32) && ((data[i]&0x7f) <128)) { fprintf(stdout,"%c",data[i]&0x7f); }
//    }
//  fprintf(stdout,"\n");
  if ((line==0)) { 
    if (thepage.valid) {
      if (thepage.pnum==the_page) {
        if (debug) { 
           fprintf(stdout,"Previous page:\n");
           debug_print_page(&prevpage);
           fprintf(stdout,"Current page:\n");
           debug_print_page(&thepage);
           fprintf(stdout,"Current page subtitle output:\n");
           print_page(&thepage);
           fprintf(stdout,"Actual subtitle output:\n");
        }
        i=bcmp(&thepage.pagebuf[40],&prevpage.pagebuf[40],24*40);
        if (debug) { fprintf(stdout,"Comparison of pages: %d\n",i); }
	if (i!=0) {
          print_page(&thepage);
          prevpage=thepage;
	}
      }
    }
    thepage.pnum=(mag<<8)|unham(data[0],data[1]); // The lower two (hex) numbers of page
    if (thepage.pnum==the_page) {
      thepage.valid=1;
    } else {
      thepage.valid=0;
    }
//    fprintf(stderr,"thepage.pnum=%04x\n",thepage.pnum);
//    if (thepage.pnum==0xff) return;  // These are filler lines. Can use to update clock
    thepage.flags=unham(data[2],data[3])&0x80;
    thepage.flags|=(c&0x40)|((c>>2)&0x20);
    c=unham(data[6],data[7]);
    thepage.flags|=((c<<4)&0x10)|((c<<2)&0x08)|(c&0x04)|((c>>1)&0x02)|((c>>4)&0x01);
    thepage.lang=((c>>5) & 0x07);
    thepage.PTS=PTS;

    if (thepage.flags&0x80) {
      thepage.num_valid_lines=0;
      for (i=1;i<25;i++) { thepage.line_transmitted[i]=0; }
      memset(thepage.pagebuf,' ', 25*40);
    }

    c=unham(data[8],data[9]);

    thepage.sub=(unham(data[4],data[5])<<8)|(unham(data[2],data[3])&0x3f7f);
  } 

  if (thepage.valid) {
    if (line <= 23) {
      thepage.line_transmitted[line]=1;
      memcpy(&thepage.pagebuf[40*line],data,40);
    }
    if (line > 0) { thepage.num_valid_lines++; }
  }
}

uint64_t get_pes_pts (unsigned char* buf) {
  uint64_t PTS;
  int PTS_DTS_flags;
  uint64_t p0,p1,p2,p3,p4;

  PTS_DTS_flags=(buf[7]&0xb0)>>6;
  if ((PTS_DTS_flags&0x02)==0x02) {
    // PTS is in bytes 9,10,11,12,13
    p0=(buf[13]&0xfe)>>1|((buf[12]&1)<<7);
    p1=(buf[12]&0xfe)>>1|((buf[11]&2)<<6);
    p2=(buf[11]&0xfc)>>2|((buf[10]&3)<<6);
    p3=(buf[10]&0xfc)>>2|((buf[9]&6)<<5);
    p4=(buf[9]&0x08)>>3;

    PTS=p0|(p1<<8)|(p2<<16)|(p3<<24)|(p4<<32);
    PTS=PTS/90;
  } else {
    PTS=0;
  }
  return(PTS);
}

int process_pes_packet (unsigned char* buf,int n, int the_page) {
  int pid;
  unsigned char mpag,mag,line;
  int i,j,k,page;
  int stream_id;
  int PES_packet_length;
  int PTS_DTS_flags;
  int data_len,data_unit_id;
  int c,d,e;
  uint64_t p0,p1,p2,p3,p4;

  stream_id=buf[3];
  PES_packet_length=(buf[4]<<8)|buf[5];
  //  fprintf(stderr,"n=%d\n",n);
  //  fprintf(stderr,"stream_id=0x%02x\n",stream_id);
  //  fprintf(stderr,"PES_packet_length=%d\n",PES_packet_length);
//  fprintf(stderr,"buf[6]=%02x\n",buf[6]);

  if ((stream_id==0xbd) && ((buf[buf[8]+9]>=0x10) && (buf[buf[8]+9]<=0x1f))) {
    PTS_DTS_flags=(buf[7]&0xb0)>>6;
    if (PTS_DTS_flags==0x02) {
      // PTS is in bytes 9,10,11,12,13
      p0=(buf[13]&0xfe)>>1|((buf[12]&1)<<7);
      p1=(buf[12]&0xfe)>>1|((buf[11]&2)<<6);
      p2=(buf[11]&0xfc)>>2|((buf[10]&3)<<6);
      p3=(buf[10]&0xfc)>>2|((buf[9]&6)<<5);
      p4=(buf[9]&0x08)>>3;
 
      PTS=p0|(p1<<8)|(p2<<16)|(p3<<24)|(p4<<32);
      PTS=PTS/90;
      if (FIRST_PTS==0) { FIRST_PTS=PTS; }
    } else {
      //fprintf(stdout,"stream_id=%02x, No PTS\n",stream_id);
      if (no_pts_warning==0) {
        fprintf(stderr,"WARNING: No PTS value in teletext packet - using audio or video PTS.\n");
        fprintf(stderr,"First audio PTS=%lld ms, First video PTS=%lld ms\n",first_audio_pts,first_video_pts);
        no_pts_warning=1;
      }
//      fprintf(stderr,"here: audio_pts=%lld ms, video_pts=%lld ms\n",audio_pts,video_pts);
      PTS=audio_pts;
      if (FIRST_PTS==0) { FIRST_PTS=first_video_pts; }
    }
    k=buf[8]+9;
//    fprintf(stderr,"PES Header Data Length=%d\n",buf[8]);
//    fprintf(stderr,"HERE: k=%d, n=%d\n",k,n);
    k++;
    while (k < n) {
      data_unit_id=buf[k++];
      data_len=buf[k++];
      if (data_len!=0x2c) { data_len=0x2c; }
//      fprintf(stdout,"data_unit_id=%02x,data_len=%d\n",data_unit_id,data_len);
//      if ((data_unit_id==0x02) || (data_unit_id==0x03)) {
        for (j=k;j<k+data_len;j++) { 
          buf[j]=invtab[buf[j]];
        }
        mpag=unham(buf[k+2],buf[k+3]);
        mag=mpag&7;
        line=(mpag>>3)&0x1f;
        // mag==0 means page is 8nn
        if (mag==0) mag=8;
        set_line(line,&buf[k+4],mag,the_page);
//        }
      k+=data_len;
    }
  } else {
     fprintf(stderr,"This is not a private data type 1 stream - are you sure you specified the correct PID? stream_id=%02x\n",stream_id);
     exit(1);
  }
}


int main(int argc, char** argv)
{
unsigned char buf[188];
unsigned char pesbuf[1265536];
int peslength;
int stream_id;
uint64_t tmp_pts;
unsigned short pid;
int c;
int pes_format=0;
ts_status_t ts_status;
int packet;
int counter;
int adaption_field_control,continuity_counter;
int adaption_field_length;
int discontinuity_indicator;
int splicing_point_flag;
int max_pes = 0;
int min_pes = 9999999;
int pes_written = 0;
int pes_dropped = 0;
int ts_dropped = 0;
int pes_dirty;
int stream_packets;
int i, j, n, m;
#ifndef WITHOUT_DVB
char pol=0;
unsigned long freq = -1;
unsigned long sr = -1;
fe_spectral_inversion_t specInv = INVERSION_AUTO;
int tone = -1;
fe_modulation_t modulation = CONSTELLATION_DEFAULT;
fe_transmit_mode_t TransmissionMode = TRANSMISSION_MODE_DEFAULT;
fe_bandwidth_t bandWidth = BANDWIDTH_DEFAULT;
fe_guard_interval_t guardInterval = GUARD_INTERVAL_DEFAULT;
fe_code_rate_t HP_CodeRate = HP_CODERATE_DEFAULT;
unsigned int diseqc = 0;
int fd_frontend;
int fd_sec;
int fd_dvr;
#endif
int x;

/* Test channels - Astra 19E, 11837h, SR 27500 - e.g. ARD */
int pids[MAX_CHANNELS];
int pnrs[MAX_CHANNELS] = { 1, 2, 3, 4, 5, 6, 7, 8 }; /* Directory names */

mag_struct mags[MAX_CHANNELS][8];
int fd[MAX_CHANNELS];
int count;
int t = 0;
int use_stdin = 1;
int hdr_len;
int the_page;
int the_pid;

is_first = 0;
 
memset(prevpage.pagebuf,' ', 25 * 40);
for(i = 1; i < 25; i++) { prevpage.line_transmitted[i] = 0; }
prevpage.num_valid_lines = 0;

fprintf(stderr, "dvbtextsubs v%s - (C) Dave Chapman 2003-2004\n", VERSION);
fprintf(stderr, "Latest version available from http://www.linuxstb.org\n");

//params.iFrequency=;
//params.SymbolRate=;
//params.FEC_inner=;

if (argc == 1)
	{
	fprintf(stderr, USAGE);
	return -1;
	}
else
	{
	count = 0;
	for(i = 1; i < argc; i++)
		{
		if(strcmp(argv[i], "-d") == 0)
			{
			debug = 1;
			}
		else if(strcmp(argv[i], "-vdr") == 0)
			{
			pes_format = 1;
#ifndef WITHOUT_DVB
			}
		else if(strcmp(argv[i], "-f") == 0)
			{
			i++;
			freq = atoi(argv[i]) * 1000UL;
			use_stdin = 0;
			}
		else if(strcmp(argv[i], "-D") == 0)
			{
			i++;
			diseqc = atoi(argv[i]);
			use_stdin = 0;
			}
		else if(strcmp(argv[i], "-p") == 0)
			{
			i++;
			if(argv[i][1] == 0)
				{
				if(tolower(argv[i][0]) == 'v')
					{
					pol = 'v';
					}
				else if(tolower(argv[i][0]) == 'h')
					{
					pol = 'h';
					}
				}
			use_stdin = 0;
			}
		else if(strcmp(argv[i], "-s") == 0)
			{
			i++;
			sr = atoi(argv[i]) * 1000UL;
			use_stdin = 0;
			}
		else if(strcmp(argv[i], "-r") == 0)
			{
    	    srt_flag = 1;
#endif
			} 
		else if(strcmp(argv[i], "-o") == 0)
			{
			i++;
    	    offset = atof(argv[i]);
			} 
		else
			{
			pids[count] = atoi(argv[i]);
			if(pids[count]) count++ ;
			}
		}
	}

  if ((pes_format==1) && (count==1)) {
    the_page=(pids[0]%10)|((((pids[0]-(100*(pids[0]/100)))%100)/10)<<4)|((pids[0]/100)<<8);
  } else if ((pes_format==0) && (count==2)) {
    the_pid=pids[0];
    the_page=(pids[1]%10)|((((pids[1]-(100*(pids[1]/100)))%100)/10)<<4)|((pids[1]/100)<<8);
  } else {
    fprintf(stderr,USAGE);
    return(-1);
  }

#ifndef WITHOUT_DVB
  if ((use_stdin==0) && (pol!=0) && (freq!=-1) && (sr!=-1)) {
    if((fd_frontend = open("/dev/dvb/adapter0/frontend0",O_RDWR)) < 0){
        fprintf(stderr,"FD %i: ",i);
        perror("FRONTEND DEVICE: ");
        return -1;
    }

    fprintf(stderr,"Tuning channel: %d%c sr %d\n",freq,pol,sr);
    tune_it(fd_frontend,fd_sec,freq,sr,pol,tone,specInv,diseqc,modulation,HP_CodeRate,TransmissionMode,guardInterval,bandWidth);
    close(fd_sec);
    close(fd_frontend);
  }
#endif

  m=0;
    mags[m][0].mag=8;
    for (i=1;i<8;i++) { 
      mags[m][i].mag=i; 
      mags[m][i].valid=0;
    }

#ifndef WITHOUT_DVB
  if (use_stdin==0) {
    for (i=0;i<1;i++) {  
      if((fd[i] = open("/dev/dvb/adapter0/demux0",O_RDWR)) < 0){
        fprintf(stderr,"FD %i: ",i);
        perror("DEMUX DEVICE: ");
        return -1;
      }
    }

    if((fd_dvr = open("/dev/dvb/adapter0/dvr0",O_RDONLY)) < 0){
      perror("DVR DEVICE: ");
      return -1;
    }
    /* Now we set the filters */
    for (i=0;i<1;i++) {
      set_tt_filt(fd[i],pids[i]);
    }
  }
#endif

  thepage.valid=0;

	if(! srt_flag)
		{
	  printf("<?xml version=\"1.0\" encoding=\"iso-8859-1\" ?>\n");
	  printf("<subpictures>\n");
	  printf("  <stream forced=\"no\">\n");
		}

  if (pes_format==1) {
    n = fread(pesbuf,1,6,stdin);
    while (n > 0) {
      if ((pesbuf[0]!=0) || (pesbuf[1]!=0) || (pesbuf[2]!=1)) {
        fprintf(stderr,"PES ERROR: does not start with 0x000001\n");
        fprintf(stderr,"%02x %02x %02x %02x %02x\n",pesbuf[0],pesbuf[1],pesbuf[2],pesbuf[3],pesbuf[4]);
        exit(0);
      }
      // stream_id: e0=video, c0=audio, bd=DVB subtitles, AC3 etc
      stream_id=pesbuf[3];
      peslength=((pesbuf[4]<<8)|pesbuf[5]);
      //      fprintf(stderr,"stream_id=%02x, PES_packet_length=%d\n",stream_id,peslength);
      n = fread(&pesbuf[6],1,peslength,stdin);
      if (stream_id==0xbd) {
        process_pes_packet(pesbuf,peslength+6,the_page);
      } else if (stream_id==0xe0) {
        tmp_pts=get_pes_pts(pesbuf);
        //fprintf(stdout,"stream_id=%02x, PTS=%lld ms\n",stream_id,tmp_pts);
	if (tmp_pts > video_pts) { video_pts=tmp_pts; }
        if (first_video_pts==0) { first_video_pts=video_pts; }
      } else if (stream_id==0xc0) {
        tmp_pts=get_pes_pts(pesbuf);
        //fprintf(stdout,"stream_id=%02x, PTS=%lld ms\n",stream_id,tmp_pts);
	if (tmp_pts > audio_pts) { audio_pts=tmp_pts; }
        if ((first_audio_pts==0) || ((tmp_pts!=0) && (tmp_pts < first_audio_pts))) { first_audio_pts=tmp_pts; }
      }
      n = fread(pesbuf,1,6,stdin);
    }
    printf("  </stream>\n");
    printf("</subpictures>\n");
    exit(0);
  } else {
    packet=0;
    stream_packets=0;
    peslength=-1;
    counter=-1;
    ts_status=TS_WAITING;
    for (;;) {
      c=0;
      while (c<188) {
        if (use_stdin) {
          n = fread(&buf[c],1,188-c,stdin);
#ifndef WITHOUT_DVB
        } else {
          n = read(fd_dvr,&buf[c],188-c);
          if (n < 0) { 
            if (errno == EAGAIN) continue;
            if (errno == EINTR) continue;
   
            perror("Error reading from fd_dvr"); exit(1);
          }
#endif
        }
        if (n==0)
			{
			
if(! srt_flag)
		{
          printf("  </stream>\n");
          printf("</subpictures>\n");
		}


          fprintf(stderr,"END OF STREAM\n");
          fprintf(stderr,"Processed %d TS packets (%d bytes).\n",packet,188*packet);
          fprintf(stderr,"Number of TS packets missing:  %d\n",ts_dropped);
          fprintf(stderr,"Minimum PES packet size:       %d bytes\n",min_pes);
  	fprintf(stderr,"Maximum PES packet size:       %d bytes\n",max_pes);
          fprintf(stderr,"Number of PES packets written: %d\n",pes_written);
          fprintf(stderr,"Number of PES packets dropped: %d\n",pes_dropped);
          exit(0);
        }
        c+=n;
      }
      if (buf[0]!=0x47) {
        fprintf(stderr,"ERROR: incomplete TS packet\n");
      }
  
      if ( (((buf[1]&0x1f)<<8) | buf[2]) == pids[0]) {
        continuity_counter=buf[3]&0x0f;
        adaption_field_control=(buf[3]&0x30)>>4;
        discontinuity_indicator=0;
        if ((adaption_field_control==2) || (adaption_field_control==3)) {
          if (buf[4] > 0) {  // adaption_field_length
            discontinuity_indicator=(buf[5]&0x80)>>7;
  //          splicing_point_flag=(buf[5]&0x04);
  //          fprintf(stderr,"packet: %d, flags=%02x\n",packet,buf[5]);
          }
        }
        if (discontinuity_indicator) {
          fprintf(stderr,"INFORMATION: packet %d - discontinuity_indicator set\n",packet);
        }
  
        /* Firstly, check the integrity of the stream */
        if (counter==-1) {
          counter=continuity_counter;
        } else {
          if ((adaption_field_control!=0) && (adaption_field_control!=2)) {
            counter++; counter%=16;
          }
        }
  
        if (counter!=continuity_counter) {
          if (discontinuity_indicator==0) {
            n=(continuity_counter+16-counter)%16;
             
  //          fprintf(stderr,"TS: missing %d packet(s), packet=%d, expecting %02x, received %02x  afc=%d, di=%d\n",n,packet,counter,continuity_counter,adaption_field_control,discontinuity_indicator);
            ts_dropped+=n;
            pes_dirty=1;
          }
          counter=continuity_counter;
        }
  
        // Check payload start indicator.
        if (buf[1]&0x40) {
  //        fprintf(stderr,"%d: payload start\n",packet);
  //        fprintf(stderr,"previous peslength=%d\n",peslength);
          if (ts_status==TS_IN_PAYLOAD) {
            if ((pes_dirty==0) && (check_pes(pesbuf,peslength))) {
              process_pes_packet(pesbuf,peslength,the_page);
  	    // fprintf(stderr,"processed PES packet - %d bytes\n",peslength);
              if (peslength > max_pes) { max_pes=peslength; }
              if (peslength < min_pes) { min_pes=peslength; }
              pes_written++;
            } else {
              fprintf(stderr,"Dropped PES packet - %d bytes - Bytes read from stream=%d\n",peslength,stream_packets*188);
              pes_dropped++;
            }
            pes_dirty=0;
          } else {
            ts_status=TS_IN_PAYLOAD;
          }
          peslength=0;
        }
  
        if (ts_status==TS_IN_PAYLOAD) {
  //        fprintf(stderr,"processing packet\n");
          i=4;
          if ((adaption_field_control==2) || (adaption_field_control==3)) {
            // Adaption field!!!
            adaption_field_length=buf[i++];
            if (adaption_field_length > 182) {
              fprintf(stderr,"WARNING: adaption field_length > TS packet size!, adaption_field_control=%d\n",adaption_field_control);
              adaption_field_length=182;
              ts_status=TS_WAITING;
              continue;
            }
            i+=adaption_field_length;
  //          fprintf(stderr,"Adaption field length=%d\n",adaption_field_length);
          }
          if ((adaption_field_control==1) || (adaption_field_control==3)) {
            // Data
  //          fprintf(stderr,"(before)peslength=%d\n",peslength);
  //          fprintf(stderr,"copying %d bytes to pesbuf\n",188-i);
            if ((peslength+(188-i)) > sizeof(pesbuf)) {
              fprintf(stderr,"ERROR: TS PACKET %d: PES packet > %d bytes, skipping end\n",packet,sizeof(pesbuf));
            } else {
              memcpy(&pesbuf[peslength],&buf[i],188-i);
            }
            peslength+=(188-i);
  //          fprintf(stderr,"(after)peslength=%d\n",peslength);
          }
  
          if (verbose) fprintf(stderr,"packet %d, counter=%d, cc=%d, afc=%d\n",packet,counter,continuity_counter,adaption_field_control);
        }
        packet++;
      }
      stream_packets++;
    }
  }
}
